/*
 * Copyright (c) 2009 The openGion Project.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package org.opengion.fukurou.util;

import java.io.File;
import java.io.StringWriter;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.Arrays;
import java.util.Map;
import java.util.WeakHashMap;

import javax.tools.JavaCompiler;
import javax.tools.StandardJavaFileManager;
import javax.tools.ToolProvider;
import javax.tools.JavaCompiler.CompilationTask;

/**
 * AutoCompile機能、HotDeploy機能を実現するためのクラスローダーです。
 *
 * AutoCompile機能は、クラスの動的コンパイルを行います。
 * AutoCompile機能を有効にするには、コンストラクタで与えられるHybsLoaderConfigオブジェクトで、
 * AutoCompileフラグをtrueにしておく必要があります。
 *
 * HotDeploy機能は、クラスの動的ロードを行います。
 * HotDeploy機能を有効にするには、コンストラクタで与えられるHybsLoaderConfigオブジェクトで、
 * HotDeployフラグをtrueにしておく必要があります。
 * 
 * (1)クラスの動的コンパイル
 *  {@link #loadClass(String)}メソッドが呼ばれた場合に、ソースディレクトより、対象となるソースファイルを
 *  検索し、クラスのコンパイルを行います。
 *  コンパイルが行われる条件は、「クラスファイルが存在しない」または「クラスファイルのタイムスタンプがソースファイルより古い」です。
 *  
 *  コンパイルを行うには、JDKに含まれるtools.jarが存在している必要があります。
 *  tools.jarが見つからない場合、エラーとなります。
 *  
 *  また、コンパイルのタスクのクラス(オブジェクト)は、JVMのシステムクラスローダー上のクラスに存在しています。
 *  このため、サーブレットコンテナで、通常読み込まれるWEB-INF/classes,WEB-INF/lib以下のクラスファイルも、
 *  そのままでは参照することができません。
 *  これらのクラスを参照する場合は、HybsLoaderConfigオブジェクトに対してクラスパスを設定しておく必要があります。
 *  
 * (2)クラスロード
 *  クラスの動的ロードは、クラスローダーの入れ替えによって実現しています。
 *  HotDeploy機能を有効にした場合、読み込むクラス単位にURLClassLoaderを生成しています。
 *  クラスロードを行う際に、URLClassLoaderを新しく生成することで、クラスの再ロードを行っています。
 *  つまり、HotDeployにより読み込まれるそれぞれのクラスは、お互いに独立した(平行な位置に存在する)関係に
 *  なります。
 *  このため、あるHotDeployによりロードされたクラスAから、同じくHotDeployによりロードされたクラスBを直接参照
 *  することができません。
 *  この場合は、クラスBのインターフェースを静的なクラスローダー(クラスAから参照できる位置)に配置することで、クラスB
 *  のオブジェクトにアクセスすることができます。
 * 
 * @og.rev 5.1.1.0 (2009/12/01) 新規作成
 * @og.group 業務ロジック
 * 
 * @version 5.0
 * @author Hiroki Nakamura
 * @since JDK1.6,
 */
public class HybsLoader {
	private static final String CR = System.getProperty("line.separator");

	// HotDeploy機能を使用しない場合のURLClossLoaderのキャッシュキー
	private static final String CONST_LOADER_KEY = "CONST_LOADER_KEY";
	
	private static final JavaCompiler COMPILER = ToolProvider.getSystemJavaCompiler();
	private static final StandardJavaFileManager FILE_MANAGER = COMPILER.getStandardFileManager(null, null, null);
	
	private final Map<String, HybsURLClassLoader> loaderMap = new WeakHashMap<String, HybsURLClassLoader>();
	private final Map<String, String> clsNameMap = new WeakHashMap<String, String>();
	private final String srcDir;
	private final String classDir;
	private final boolean isHotDeploy;
	private final boolean isAutoCompile;
	private final String classPath;

	/**
	 * HybsLoaderOptionを使用してHybsLoaderオブジェクトを生成します。
	 * 
	 * @param option HybsLoaderを構築するための設定情報
	 */
	public HybsLoader( final HybsLoaderConfig option ) {
		srcDir = option.getSrcDir();
		classDir = option.getClassDir();
		isHotDeploy = option.isHotDeploy();
		isAutoCompile = option.isAutoCompile();
		classPath = option.getClassPath();
	}

	/**
	 * 指定されたクラス名のクラスをロードします。
	 * クラス名については、クラス自身の名称のみを指定することができます。
	 * (パッケージ名を含めた完全な形のクラス名を指定することもできます)
	 * 
	 * @param clsNm クラス名
	 * @return クラス
	 */
	public Class<?> load( final String clsNm ) {
		String clsName = getQualifiedName( clsNm );
		if( isAutoCompile ) {
			compileClass( clsName );
		}
		Class<?> cls = loadClass( clsName );

		return cls;
	}

	/**
	 * 指定されたクラス名のクラスをロードし、デフォルトコンストラクターを使用して
	 * インスタンスを生成します。
	 * 
	 * @og.rev 5.1.8.0 (2010/07/01) Exceptionのエラーメッセージの修正(状態の出力)
	 * 
	 * @param clsName クラス名(Qualified Name)
	 * @return インスタンス
	 */
	public Object newInstance( final String clsName ) {
		Class<?> cls = load( clsName );
		Object obj = null;
		try {
			obj = cls.newInstance();
		}
		catch( InstantiationException ex ) {
//			throw new RuntimeException( "インスタンスの生成に失敗しました" );
			String errMsg = "インスタンスの生成に失敗しました。["  + clsName + "]" ;
			throw new RuntimeException( errMsg , ex );
		}
		catch( IllegalAccessException ex ) {
//			throw new RuntimeException( "アクセスが拒否されました" );
			String errMsg = "アクセスが拒否されました。["  + clsName + "]" ;
			throw new RuntimeException( errMsg , ex );
		}
		return obj;
	}

	/**
	 * クラス名より完全クラス名を検索します。
	 * 
	 * @param clsNm クラス名
	 * @return 完全クラス名
	 */
	private String getQualifiedName( final String clsNm ) {
		String clsName = null;
		if( clsNm.indexOf( '.' ) >= 0 ) {
			clsName = clsNm;
		}
		else {
			synchronized( clsNameMap ) {
				clsName = clsNameMap.get( clsNm );
				if( clsName == null ) {
					clsName = findFile( "", clsNm );
				}
				clsNameMap.put( clsNm, clsName );
			}
			if( clsName == null ) {
				throw new RuntimeException( "ソースファイルが存在しません。ファイル=[" + clsNm + "]" );
			}
		}
		return clsName;
	}

	/**
	 * クラス名に対応するJavaファイルを再帰的に検索します。
	 * 
	 * @param path 既定パス
	 * @param nm クラス名
	 * @return 完全クラス名
	 */
	private String findFile( final String path, final String nm ) {
		String tmpSrcPath = srcDir + path;
		File[] files = (new File( tmpSrcPath )).listFiles();
		if( files != null && files.length > 0 ) {
			for( int i=0; i<files.length; i++ ) {
				if ( files[i].isDirectory() ) {
					String rtn = findFile( path + files[i].getName() + File.separator, nm );
					if( rtn != null && rtn.length() > 0 ) {
						return rtn;
					}
				}
				else if( ( nm + ".java" ).equals( files[i].getName() ) ) {
					return path.replace( File.separatorChar, '.' ) + nm;
				}
			}
		}
		return null;
	}

	/**
	 * クラスをコンパイルします。
	 * 
	 * @og.rev 5.1.8.0 (2010/07/01) ソースファイルのエンコードは、UTF-8にする。
	 * 
	 * @param clsNm クラス名
	 */
	private void compileClass( final String clsNm ) {
		if( COMPILER == null ) {
			throw new RuntimeException( "コンパイラクラスが定義されていません。tools.jarが存在しない可能性があります" );
		}
		
		String srcFqn = srcDir + clsNm.replace( ".", File.separator ) + ".java";

		File srcFile = new File( srcFqn );
		if( !srcFile.exists() ) {
			throw new RuntimeException( "ソースファイルが存在しません。ファイル=[" + srcFqn + "]" );
		}

		String classFqn =  classDir + clsNm.replace( ".", File.separator ) + ".class";
		File clsFile = new File ( classFqn );
		if( clsFile.exists() && srcFile.lastModified() <= clsFile.lastModified() ) {
			return;
		}

		if( !clsFile.getParentFile().exists() ) {
			if( !clsFile.getParentFile().mkdirs() ) {
				throw new RuntimeException( "ディレクトリが作成できませんでした。ファイル=[" + clsFile + "]" );
			}
		}

		StringWriter sw = new StringWriter();
		File[] sourceFiles = { new File( srcFqn ) };
		// 5.1.8.0 (2010/07/01) ソースファイルのエンコードは、UTF-8にする。
//		String[] cpOpts = new String[]{ "-d", classDir, "-classpath", classPath };
		String[] cpOpts = new String[]{ "-d", classDir, "-classpath", classPath, "-encoding", "UTF-8" };

		CompilationTask task = COMPILER.getTask(sw, FILE_MANAGER, null, Arrays
				.asList(cpOpts), null, FILE_MANAGER
				.getJavaFileObjects(sourceFiles));

		boolean isOk = false;
		// lockしておかないと、java.lang.IllegalStateExceptionが発生することがある
		synchronized( this ) {
			isOk = task.call();
		}
		if( !isOk ) {
			throw new RuntimeException( "コンパイルに失敗しました。" + CR + sw.toString() );
		}
	}

	/**
	 * クラスをロードします。
	 * 
	 * @param clsNm クラス名
	 */
	private Class<?> loadClass( final String clsNm ) {
		String key = isHotDeploy ? clsNm : CONST_LOADER_KEY;

		String classFqn =  classDir + clsNm.replace( ".", File.separator ) + ".class";
		File clsFile = new File( classFqn );
		if( !clsFile.exists() ) {
			throw new RuntimeException( "クラスファイルが存在しません。ファイル=[" + classFqn + "]" );
		}
		long lastClassModifyTime = clsFile.lastModified();

		HybsURLClassLoader loader = null;
		synchronized( loaderMap ) {
			loader = loaderMap.get( key );
			if( loader == null || lastClassModifyTime > loader.getCreationTime() ) {
				try {
					loader = new HybsURLClassLoader( new URL[] { ( new File( classDir ).toURI().toURL() ) }, this.getClass().getClassLoader() );
				}
				catch( MalformedURLException ex ) {
					throw new RuntimeException( "クラスロードのURL変換に失敗しました。ファイル=[" + classFqn + "]", ex );
				}
				loaderMap.put( key, loader );
			}
		}

		Class<?> cls;
		try {
			cls = loader.loadClass( clsNm );
		}
		catch( ClassNotFoundException ex ) {
//			throw new RuntimeException( "クラスが存在しません。ファイル=[" + classFqn + "]" );
			String errMsg = "クラスが存在しません。ファイル=[" + classFqn + "]"  ;
			throw new RuntimeException( errMsg , ex );
		}
		return cls;
	}

	/**
	 * URLClassLoaderを拡張し、クラスローダーの生成時間を管理できるようにしています。
	 */
	private static class HybsURLClassLoader {
		final URLClassLoader loader;
		final long creationTime;

		HybsURLClassLoader( final URL[] urls, final ClassLoader clsLd ) {
			loader = new URLClassLoader( urls, clsLd );
			creationTime = System.currentTimeMillis();
		}

		HybsURLClassLoader( final URL[] urls ) {
			this( urls, null );
		}

		Class<?> loadClass( final String clsName ) throws ClassNotFoundException {
			return loader.loadClass( clsName );
		}

		long getCreationTime() {
			return creationTime;
		}
	}
}
