/*
 * Copyright (c) 2009 The openGion Project.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package org.opengion.hayabusa.taglib;

import org.opengion.hayabusa.common.HybsSystem;
import org.opengion.hayabusa.common.HybsSystemException;
import org.opengion.fukurou.util.HybsEntry;

import org.opengion.fukurou.util.StringUtil ;
import static org.opengion.fukurou.util.StringUtil.nval ;

/**
 * WriteTableTag にパラメーターを渡す為のスーパークラスです。
 *
 * writeTable タグに対して、EXEC_SQL 情報や、tableName情報を付加することができます。
 * EXEC_SQL 情報とは、タブ区切りファイルやXML形式ファイルの先頭(key="First")
 * または、最後(key="Last")に、SQL文を記述することで、ファイル取り込み時の
 * 前処理、後処理を処理する為の情報です。
 * tableName情報は、XMLファイルのROWSET属性にセットすることで、XMLファイルの登録テーブル名を
 * 指定することができます。
 *
 * この情報は、複数件登録できるため、通常の writeTable タグに属性を追加すると、
 * 複雑になるため、複数登録できる用に、内部にタグを持てる構造にします。
 *
 * @og.formSample
 * ●形式：
 *     &lt;og:writeTableParam key="[First|Last|TableName]" &gt;
 *          &lt;jsp:text&gt;delete from GE12 where SYSTEM_ID='**' and KBSAKU='0'&lt;/jsp:text&gt;
 *     &lt;/og:writeTableParam
 *
 * ●body：あり
 *
 * ●使用例
 *     &lt;og:writeTable ･････ &gt;
 *         &lt;og:writeTableParam
 *             key  = "Tablename"  value="GE12"
 *         /&gt;
 *         &lt;og:writeTableParam
 *             key  = "First"             First:最初に登録
 *         &gt;
 *              &lt;jsp:text&gt;
 *                      insert into GE12bk
 *                         select * from GE12
 *                         where SYSTEM_ID='**'
 *              &lt;/jsp:text&gt;
 *         &lt;/og:writeTableParam
 *         &lt;og:writeTableParam
 *             key  = "First"             First:の２番目に登録
 *         &gt;
 *              &lt;jsp:text&gt;delete from GE12 where SYSTEM_ID='**' and KBSAKU='0'&lt;/jsp:text&gt;
 *         &lt;/og:writeTableParam
 *         &lt;og:writeTableParam
 *             key  = "Last"              Last:最後に登録
 *         &gt;
 *              &lt;jsp:text&gt;update GE12 set XXXX='YYYY' where SYSTEM_ID='**' and KBSAKU='0'&lt;/jsp:text&gt;
 *         &lt;/og:writeTableParam
 *     &lt;/og:writeTableParam
 *
 * @og.rev 4.0.0 (2005/01/31) 新規作成
 * @og.group ファイル出力
 *
 * @version  4.0
 * @author   Kazuhiko Hasegawa
 * @since    JDK5.0,
 */
public class WriteTableParamTag extends CommonTagSupport {
	//* このプログラムのVERSION文字列を設定します。	{@value} */
	private static final String VERSION = "4.0.0 (2005/08/31)" ;

	private static final long serialVersionUID = 4000 ;	// 4.0.0 (2005/02/28)

//	private static final String[] KEY_LIST = new String[] { "First","Last","TableName" };
	private static final String[] KEY_LIST = new String[] { "First","Last","TableName","Size" }; // 5.1.6.0 (2010/05/01)

	private String key		= null;		// "First","Last","TableName" のどれか
	private String value	= null ;	// 実行するSQL文字列

	/**
	 * Taglibの開始タグが見つかったときに処理する doStartTag() を オーバーライドします。
	 *
	 * @return	int
	 */
	@Override
	public int doStartTag() {
		if( value != null ) {
			return(SKIP_BODY);				// Body を評価しない
		}
		else {
			return( EVAL_BODY_BUFFERED );	// Body を評価する。（ extends BodyTagSupport 時）
		}
	}

	/**
	 * Taglibのタグ本体を処理する doAfterBody() を オーバーライドします。
	 *
	 * @return	int
	 */
	@Override
	public int doAfterBody() {
		value = getBodyString().trim();
		return(SKIP_BODY);
	}

	/**
	 * Taglibの終了タグが見つかったときに処理する doEndTag() を オーバーライドします。
	 *
	 * @return	int
	 */
	@Override
	public int doEndTag() {
		debugPrint();		// 4.0.0 (2005/02/28)

		WriteTableTag writeTable = (WriteTableTag)findAncestorWithClass( this,WriteTableTag.class );
		if( writeTable == null ) {
			String errMsg = "<b>このタグは、WriteTableTagの内側(要素)に記述してください。</b>";
			throw new HybsSystemException( errMsg );
		}

		writeTable.addParam( new HybsEntry( key,value ) );

		return(EVAL_PAGE);
	}

	/**
	 * タグリブオブジェクトをリリースします。
	 * キャッシュされて再利用されるので、フィールドの初期設定を行います。
	 *
	 */
	@Override
	protected void release2() {
		super.release2();		// 3.5.6.0 (2004/06/18) 追加(抜けていた)
		key		= null;		// "First","Last","TableName" のどれか
		value	= null ;	// 実行するSQL文字列
	}

	/**
	 * 【TAG】パラメータとして渡すキー情報([First|Last|TableName])を指定しま。
	 *
	 * @og.tag
	 * WriteTable に渡すパラメータのキー情報です。
	 * First と Last を指定した場合は、XML 形式で出力する EXEC_SQL 情報になります。
	 * TableName の場合は、XML 形式の ROWSET の table 属性になります。
	 * 値は、value 属性で指定するか、なければ BODY 部に記述します。
	 *
	 * @param   prmKey パラメータとして渡すキー情報([First|Last|TableName])
	 */
	public void setKey( final String prmKey ) {
		key = nval( getRequestParameter( prmKey ),key );
		if( ! check( key, KEY_LIST ) ) {
			String errMsg = "指定の値は、セットできません。key=[" + key + "]" + HybsSystem.CR
					+ "次の中から指定してください。（" + StringUtil.array2csv( KEY_LIST ) + ")" ;
			throw new HybsSystemException( errMsg );
		}
	}

	/**
	 * 【TAG】パラメータとして渡す設定値を指定します(初期値:null)。
	 *
	 * @og.tag
	 * WriteTable に渡すパラメータの設定値です。
	 * First と Last を指定した場合は、XML 形式で出力する EXEC_SQL 情報になります。
	 * TableName の場合は、XML 形式の ROWSET の table 属性になります。
	 * 値は、value 属性で指定するか、なければ BODY 部に記述します。
	 * BODY 部に記述された場合は、文字列を trim() します。
	 * 設定値は、value 属性が優先です。ここの値が、null の場合は、
	 * BODY 要素を値として使用します。
	 *
	 * @param   val パラメータとして渡す設定値
	 */
	public void setValue( final String val ) {
		value = nval( getRequestParameter( val ),value );
	}

	/**
	 * このオブジェクトの文字列表現を返します。
	 * 基本的にデバッグ目的に使用します。
	 *
	 * @return このクラスの文字列表現
	 */
	public String toString() {
		return org.opengion.fukurou.util.ToString.title( this.getClass().getName() )
				.println( "VERSION"		,VERSION	)
				.println( "key"			,key		)
				.println( "value"		,value		)
				.println( "KEY_LIST"	,KEY_LIST	)
				.println( "Other..."	,getAttributes().getAttribute() )
				.fixForm().toString() ;
	}
}
